@@ -1,21 +1,24 @@
 /*
- * Miscellaneous routines for Mathomatic.
+ * Standard routines for Mathomatic.
  *
- * Copyright (C) 1987-2010 George Gesslein II.
+ * Copyright (C) 1987-2012 George Gesslein II.
  
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public 
-    License as published by the Free Software Foundation; either 
-    version 2.1 of the License, or (at your option) any later version.
+This library is free software; you can redistribute it and/or
+modify it under the terms of the GNU Lesser General Public
+License as published by the Free Software Foundation; either
+version 2.1 of the License, or (at your option) any later version.
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of 
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
- 
-    You should have received a copy of the GNU Lesser General Public 
-    License along with this library; if not, write to the Free Software 
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+This library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+Lesser General Public License for more details.
+
+You should have received a copy of the GNU Lesser General Public
+License along with this library; if not, write to the Free Software
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+
+The chief copyright holder can be contacted at gesslein@mathomatic.org, or
+George Gesslein II, P.O. Box 224, Lansing, NY  14882-0224  USA.
  
  */
 
@@ -30,29 +33,43 @@
  */
 void
 error(str)
-const char	*str;
+const char	*str;		/* constant string to display */
 {
-	error_str = str;	/* must be a constant string, temporary strings don't work */
-#if	!SILENT
+	error_str = str;	/* save reference to str, must be a constant string, temporary strings don't work */
+#if	!SILENT && !LIBRARY
 	set_color(2);		/* set color to red */
 	printf("%s\n", str);
 	default_color();	/* restore to default color */
 #endif
 }
 
+/*
+ * Reset last call to error(), as if it didn't happen.
+ */
+void
+reset_error(void)
+{
+#if	!SILENT && !LIBRARY
+	if (error_str)
+		printf(_("Forgetting previous error.\n"));
+#endif
+	error_str = NULL;
+}
+
 /*
  * Standard function to report a warning to the user.
+ * A warning is less serious than an error.
  */
 void
 warning(str)
-const char	*str;
+const char	*str;		/* constant string to display */
 {
-	warning_str = str;	/* must be a constant string, temporary strings don't work */
-#if	!SILENT
-	if (debug_level >= 0) {
-		set_color(1);	/* set color to yellow */
+	warning_str = str;	/* save reference to str, must be a constant string, temporary strings don't work */
+#if	!SILENT && !LIBRARY
+	if (debug_level >= -1) {
+		set_color(1);		/* set color to yellow */
 		printf("Warning: %s\n", str);
-		default_color();
+		default_color();	/* restore to default color */
 	}
 #endif
 }
@@ -63,7 +80,7 @@ const char	*str;
  * There is no return.
  */
 void
-error_huge()
+error_huge(void)
 {
 	longjmp(jmp_save, 14);
 }
@@ -77,13 +94,16 @@ void
 error_bug(str)
 const char	*str;	/* constant string to display */
 {
-	error(str);		/* Display the passed error message. */
-#if	SILENT
+/* Return and display the passed error message in str. */
+	error(str);	/* str must be a constant string, temporary strings don't work */
+#if	SILENT || LIBRARY
 	printf("%s\n", str);
 #endif
 	printf(_("Please report this bug to the maintainers,\n"));
 	printf(_("along with the entry sequence that caused it.\n"));
+#if	!LIBRARY
 	printf(_("Type \"help bugs\" for info on how to report bugs found in this program.\n"));
+#endif
 	longjmp(jmp_save, 13);	/* Abort the current operation with the critical error number 13. */
 }
 
@@ -93,7 +113,7 @@ const char	*str;	/* constant string to display */
  * There is no return if an error message is displayed.
  */
 void
-check_err()
+check_err(void)
 {
 	switch (errno) {
 	case EDOM:
@@ -107,21 +127,23 @@ check_err()
 		break;
 	case ERANGE:
 		errno = 0;
-		error(_("Overflow error in constant."));
+		error(_("Floating point constant out of range."));
 		longjmp(jmp_save, 2);
 		break;
 	}
 }
 
-#if	UNIX || CYGWIN
 /*
- * Get the screen (window) width and height from the operating system.
+ * Get the current screen (window) width and height from the operating system.
  *
- * Sets the global integers screen_columns and screen_rows.
+ * Return true if the global integers screen_columns and/or screen_rows were set.
  */
-void
-get_screen_size()
+int
+get_screen_size(void)
 {
+	int	rv = false;
+
+#if	UNIX || CYGWIN
 	struct winsize	ws;
 
 	ws.ws_col = 0;
@@ -129,13 +151,52 @@ get_screen_size()
 	if (ioctl(1, TIOCGWINSZ, &ws) >= 0) {
 		if (ws.ws_col > 0) {
 			screen_columns = ws.ws_col;
+			rv = true;
 		}
 		if (ws.ws_row > 0) {
 			screen_rows = ws.ws_row;
+			rv = true;
 		}
 	}
-}
+#else
+	screen_columns = STANDARD_SCREEN_COLUMNS;
+	screen_rows = STANDARD_SCREEN_ROWS;
+	rv = true;
 #endif
+	return rv;
+}
+
+/*
+ * Allocate the display lines in the vscreen[] array.
+ * Call this before using vscreen[].
+ *
+ * Return true with vscreen[] allocated to TEXT_ROWS*current_columns characters if successful.
+ */
+int
+malloc_vscreen(void)
+{
+	int	i;
+
+	if (current_columns == 0 || ((screen_columns > 0) ? (current_columns != screen_columns) : (current_columns != TEXT_COLUMNS))) {
+		if (screen_columns > 0) {
+			current_columns = screen_columns;
+		} else {
+			current_columns = TEXT_COLUMNS;
+		}
+		for (i = 0; i < TEXT_ROWS; i++) {
+			if (vscreen[i]) {
+				free(vscreen[i]);
+			}
+			vscreen[i] = malloc(current_columns + 1);
+			if (vscreen[i] == NULL) {
+				error(_("Out of memory (can't malloc(3))."));
+				current_columns = 0;
+				return false;
+			}
+		}
+	}
+	return true;
+}
 
 /*
  * Allocate the needed global expression storage arrays.
@@ -149,8 +210,10 @@ get_screen_size()
  * Returns true if successful, otherwise Mathomatic cannot be used.
  */
 int
-init_mem()
+init_mem(void)
 {
+	if (n_tokens <= 0)
+		return false;
 	if ((scratch = (token_type *) malloc(((n_tokens * 3) / 2) * sizeof(token_type))) == NULL
 	    || (tes = (token_type *) malloc(n_tokens * sizeof(token_type))) == NULL
 	    || (tlhs = (token_type *) malloc(n_tokens * sizeof(token_type))) == NULL
@@ -164,20 +227,20 @@ init_mem()
 	return true;
 }
 
-#if	0
+#if	LIBRARY || VALGRIND
 /*
- * Free the global expression storage arrays.
+ * Free the global expression storage arrays and other known memory buffers.
  * After calling this, memory usage is reset and Mathomatic becomes unusable,
  * so do not call unless finished with using the Mathomatic code.
  *
- * This routine is not needed, because when a program exits,
+ * This routine is usually not needed, because when a program exits,
  * all the memory it allocated is released by the operating system.
- * Inclusion of this routine was requested by Tam Hanna.
+ * Inclusion of this routine was requested by Tam Hanna for use with Symbian OS.
  */
 void
-free_mem()
+free_mem(void)
 {
-	int i;
+	int	i;
 
 	clear_all();
 
@@ -197,15 +260,55 @@ free_mem()
 		}
 	}
 	n_equations = 0;
+
+	for (i = 0; i < TEXT_ROWS; i++) {
+		if (vscreen[i]) {
+			free(vscreen[i]);
+			vscreen[i] = NULL;
+		}
+	}
+	current_columns = 0;
+}
+#endif
+
+#if	DEBUG
+/*
+ * Use this function to check for any erroneous global conditions.
+ * Use of this function is rather paranoid, but helpful.
+ *
+ * Always returns true, or doesn't return on error.
+ */
+int
+check_gvars(void)
+{
+	if (!(domain_check == false &&
+	high_prec == false &&
+	partial_flag == true &&
+	symb_flag == false &&
+	sign_cmp_flag == false &&
+	approximate_roots == false &&
+	repeat_flag == false))
+		error_bug("Global vars got changed!");
+
+	if (!(zero_token.level == 1 &&
+	zero_token.kind == CONSTANT &&
+	zero_token.token.constant == 0.0 &&
+	one_token.level == 1 &&
+	one_token.kind == CONSTANT &&
+	one_token.token.constant == 1.0))
+		error_bug("Global constants got changed!");
+
+	return true;
 }
 #endif
 
 /*
  * Initialize some important global variables to their defaults.
  * This is called on startup and by process() to reset the global flags to the default state.
+ * This is also called when processing is aborted with a longjmp(3).
  */
 void
-init_gvars()
+init_gvars(void)
 {
 	domain_check = false;
 	high_prec = false;
@@ -230,7 +333,7 @@ init_gvars()
  * Clean up when processing is unexpectedly interrupted or terminated.
  */
 void
-clean_up()
+clean_up(void)
 {
 	int	i;
 
@@ -242,6 +345,7 @@ clean_up()
 #endif
 		gfp = default_out;
 	}
+	gfp_filename = NULL;
 	for (i = 0; i < n_equations; i++) {
 		if (n_lhs[i] <= 0) {
 			n_lhs[i] = 0;
@@ -255,7 +359,7 @@ clean_up()
  * so that the next sign variables returned by next_sign() will be unique.
  */
 void
-set_sign_array()
+set_sign_array(void)
 {
 	int	i, j;
 
@@ -306,7 +410,7 @@ long	*vp;
  * Similar to a restart.
  */
 void
-clear_all()
+clear_all(void)
 {
 	int	i;
 
@@ -326,16 +430,21 @@ clear_all()
 }
 
 /*
- * Return true if the specified equation space is available, allocating if necessary.
+ * Return true if the specified equation space is available,
+ * zeroing and allocating if necessary.
  */
 int
 alloc_espace(i)
 int	i;	/* equation space number */
 {
-	if (i >= N_EQUATIONS)
+	if (i < 0 || i >= N_EQUATIONS)
 		return false;
+	n_lhs[i] = 0;
+	n_rhs[i] = 0;
 	if (lhs[i] && rhs[i])
 		return true;	/* already allocated */
+	if (lhs[i] || rhs[i])
+		return false;	/* something is wrong */
 	lhs[i] = (token_type *) malloc(n_tokens * sizeof(token_type));
 	if (lhs[i] == NULL)
 		return false;
@@ -348,13 +457,38 @@ int	i;	/* equation space number */
 	return true;
 }
 
+/*
+ * Allocate all equation spaces up to and including an equation space number,
+ * making sure the specified equation number is valid and usable.
+ *
+ * Returns true if successful.
+ */
+int
+alloc_to_espace(en)
+int	en;	/* equation space number */
+{
+	if (en < 0 || en >= N_EQUATIONS)
+		return false;
+	for (;;) {
+		if (en < n_equations)
+			return true;
+		if (n_equations >= N_EQUATIONS)
+			return false;
+		if (!alloc_espace(n_equations)) {
+			warning(_("Memory is exhausted."));
+			return false;
+		}
+		n_equations++;
+	}
+}
+
 /*
  * Allocate or reuse an empty equation space.
  *
  * Returns empty equation space number ready for use or -1 on error.
  */
 int
-alloc_next_espace()
+alloc_next_espace(void)
 {
 	int	i, n;
 
@@ -387,7 +521,7 @@ alloc_next_espace()
  * Return the number of the next empty equation space, otherwise don't return.
  */
 int
-next_espace()
+next_espace(void)
 {
 	int	i;
 
@@ -404,6 +538,7 @@ next_espace()
 
 /*
  * Copy equation space "src" to equation space "dest".
+ * "dest" is overwritten.
  */
 void
 copy_espace(src, dest)
@@ -420,12 +555,14 @@ int	src, dest;	/* equation space numbers */
 }
 
 /*
- * Return true if equation space "i" is solved for a normal variable.
+ * Return true if equation space "i" is a valid equation solved for a normal variable.
  */
 int
 solved_equation(i)
 int	i;
 {
+	if (empty_equation_space(i))
+		return false;
 	if (n_rhs[i] <= 0)
 		return false;
 	if (n_lhs[i] != 1 || lhs[i][0].kind != VARIABLE || (lhs[i][0].token.variable & VAR_MASK) <= SIGN)
@@ -465,7 +602,7 @@ var_in_equation(i, v)
 int	i;	/* equation space number */
 long	v;	/* standard Mathomatic variable */
 {
-	if (n_lhs[i] <= 0)
+	if (empty_equation_space(i))
 		return false;
 	if (found_var(lhs[i], n_lhs[i], v))
 		return true;
@@ -477,9 +614,73 @@ long	v;	/* standard Mathomatic variable */
 }
 
 /*
- * Substitute every instance of "v" in "equation" with "expression".
+ * Return true if variable "v" exists in any equation space.
+ *
+ * Search forward starting at the next equation space if forward_direction is true,
+ * otherwise search backwards starting at the previous equation space.
+ * If found, return true with cur_equation set to the equation space the variable is found in.
+ */
+int
+search_all_for_var(v, forward_direction)
+long	v;
+int	forward_direction;
+{
+	int	i, n;
+
+	i = cur_equation;
+	for (n = 0; n < n_equations; n++) {
+		if (forward_direction) {
+			if (i >= (n_equations - 1))
+				i = 0;
+			else
+				i++;
+		} else {
+			if (i <= 0)
+				i = n_equations - 1;
+			else
+				i--;
+		}
+		if (var_in_equation(i, v)) {
+			cur_equation = i;
+			return true;
+		}
+	}
+	return false;
+}
+
+/*
+ * Replace all occurrences of variable from_v with to_v in an equation space.
  */
 void
+rename_var_in_es(en, from_v, to_v)
+int	en;	 	/* equation space number */
+long	from_v, to_v;	/* Mathomatic variables */
+{
+	int	i;
+
+	if (empty_equation_space(en)) {
+#if	DEBUG
+		error_bug("Invalid or empty equation number given to rename_var_in_es().");
+#else
+		return;
+#endif
+	}
+	for (i = 0; i < n_lhs[en]; i += 2)
+		if (lhs[en][i].kind == VARIABLE
+		    && lhs[en][i].token.variable == from_v)
+			lhs[en][i].token.variable = to_v;
+	for (i = 0; i < n_rhs[en]; i += 2)
+		if (rhs[en][i].kind == VARIABLE
+		    && rhs[en][i].token.variable == from_v)
+			rhs[en][i].token.variable = to_v;
+}
+
+/*
+ * Substitute every instance of "v" in "equation" with "expression".
+ *
+ * Return true if something was substituted.
+ */
+int
 subst_var_with_exp(equation, np, expression, len, v)
 token_type	*equation;	/* equation side pointer */
 int		*np;		/* pointer to equation side length */
@@ -489,9 +690,10 @@ long		v;		/* variable to substitute with expression */
 {
 	int	j, k;
 	int	level;
+	int	substituted = false;
 
 	if (v == 0 || len <= 0)
-		return;
+		return false;
 	for (j = *np - 1; j >= 0; j--) {
 		if (equation[j].kind == VARIABLE && equation[j].token.variable == v) {
 			level = equation[j].level;
@@ -503,12 +705,19 @@ long		v;		/* variable to substitute with expression */
 			blt(&equation[j], expression, len * sizeof(token_type));
 			for (k = j; k < j + len; k++)
 				equation[k].level += level;
+			substituted = true;
 		}
 	}
+	if (substituted) {
+		if (is_integer_var(v) && !is_integer_expr(expression, len)) {
+			warning(_("Substituting integer variable with non-integer expression."));
+		}
+	}
+	return substituted;
 }
 
 /*
- * Return the minimum level encountered in a Mathomatic "expression".
+ * Return the base (minimum) parentheses level encountered in a Mathomatic "expression".
  */
 int
 min_level(expression, n)
@@ -518,12 +727,19 @@ int		n;		/* expression length */
 	int		min1;
 	token_type	*p1, *ep;
 
-	if (n <= 1) {
-		if (n <= 0) {
-			return 1;
-		} else {
-			return expression[0].level;
-		}
+#if	DEBUG
+	if (expression == NULL)
+		error_bug("NULL pointer passed to min_level().");
+#endif
+	switch (n) {
+	case 1:
+		return expression[0].level;
+	case 3:
+		return expression[1].level;
+	default:
+		if (n <= 0 || (n & 1) != 1)
+			error_bug("Invalid expression length in call to min_level().");
+		break;
 	}
 	min1 = expression[1].level;
 	ep = &expression[n];
@@ -577,15 +793,17 @@ int		*np;		/* pointer to the returned parsed expression length */
 #if	LIBRARY
 	snprintf(buf, sizeof(buf), "#%+d", pull_number);
 	pull_number++;
-	cp = parse_expr(equation, np, buf);
+	cp = parse_expr(equation, np, buf, true);
+	if (extra_characters(cp))
+		return false;
 	return(cp && *np > 0);
 #else
 	for (;;) {
 		if ((cp = get_string(buf, sizeof(buf))) == NULL) {
 			return false;
 		}
-		cp = parse_expr(equation, np, cp);
-		if (cp) {
+		cp = parse_expr(equation, np, cp, true);
+		if (cp && !extra_characters(cp)) {
 			break;
 		}
 	}
@@ -633,7 +851,11 @@ int	i;	/* equation space number */
 		return true;
 	}
 	if (n_lhs[i] <= 0) {
-		error(_("Equation space is empty."));
+		if (i == cur_equation) {
+			error(_("Current equation space is empty."));
+		} else {
+			error(_("Equation space is empty."));
+		}
 		return true;
 	}
 	return false;
@@ -642,15 +864,19 @@ int	i;	/* equation space number */
 /*
  * Verify that a current equation or expression is loaded.
  *
- * Return true and display a message if current equation space is empty.
+ * Return true and display a message if the current equation space is empty or not defined.
  */
 int
-current_not_defined()
+current_not_defined(void)
 {
 	int	i;
 
 	i = cur_equation;
-	if (i < 0 || i >= n_equations || n_lhs[i] <= 0) {
+	if (i < 0 || i >= n_equations) {
+		error(_("Current equation number out of range; reset to 1."));
+		i = cur_equation = 0;
+	}
+	if (n_lhs[i] <= 0) {
 		error(_("No current equation or expression."));
 		return true;
 	}
@@ -670,7 +896,7 @@ char	*string;	/* storage for input string */
 int	n;		/* maximum size of "string" in bytes */
 {
 #if	LIBRARY
-	error(_("Missing command line argument."));
+	error(_("Library usage error. Input requested, possibly due to missing command-line argument."));
 	return NULL;
 #else
 	int	i;
@@ -678,12 +904,17 @@ int	n;		/* maximum size of "string" in bytes */
 	char	*cp;
 #endif
 
+#if	DEBUG
+	if (string == NULL)
+		error_bug("NULL pointer passed to get_string().");
+#endif
 	if (quiet_mode) {
 		prompt_str[0] = '\0';	/* don't display a prompt */
 	}
 	input_column = strlen(prompt_str);
+	fflush(NULL);	/* flush everything before gathering input */
 #if	READLINE
-	if (readline_enabled && !echo_input) {
+	if (readline_enabled) {
 		cp = readline(prompt_str);
 		if (cp == NULL) {
 			if (!quiet_mode)
@@ -698,9 +929,8 @@ int	n;		/* maximum size of "string" in bytes */
 			free(cp);
 		}
 	} else {
-		if (!echo_input) {
-			printf("%s", prompt_str);
-		}
+		printf("%s", prompt_str);
+		fflush(stdout);
 		if (fgets(string, n, stdin) == NULL) {
 			if (!quiet_mode)
 				printf(_("\nEnd of input.\n"));
@@ -708,21 +938,24 @@ int	n;		/* maximum size of "string" in bytes */
 		}
 	}
 #else
-	if (!echo_input) {
-		printf("%s", prompt_str);
-	}
+	printf("%s", prompt_str);
+	fflush(stdout);
 	if (fgets(string, n, stdin) == NULL) {
 		if (!quiet_mode)
 			printf(_("\nEnd of input.\n"));
 		exit_program(0);
 	}
 #endif
+	if (abort_flag) {
+		abort_flag = false;
+		longjmp(jmp_save, 13);
+	}
 	/* Fix an fgets() peculiarity: */
 	i = strlen(string) - 1;
 	if (i >= 0 && string[i] == '\n') {
 		string[i] = '\0';
 	}
-	if ((gfp != stdout && gfp != stderr) || echo_input) {
+	if ((gfp != stdout && gfp != stderr) || (echo_input && !quiet_mode)) {
 		/* Input that is prompted for is now included in the redirected output
 		   of a command to a file, making redirection like logging. */
 		fprintf(gfp, "%s%s\n", prompt_str, string);
@@ -737,13 +970,17 @@ int	n;		/* maximum size of "string" in bytes */
  * Display the prompt in prompt_str[] and wait for "y" or "n" followed by Enter.
  *
  * Return true if "y".
+ * Return false if "n" or not interactive.
  */
 int
-get_yes_no()
+get_yes_no(void)
 {
 	char	*cp;
 	char	buf[MAX_CMD_LEN];
 
+	if (!isatty(0)) {
+		return false;
+	}
 	for (;;) {
 		if ((cp = get_string(buf, sizeof(buf))) == NULL) {
 			return false;
@@ -769,16 +1006,26 @@ int
 return_result(en)
 int	en;	/* equation space number the result is in */
 {
-	if (n_lhs[en] <= 0) {
+	if (empty_equation_space(en)) {
 		return false;
 	}
 #if	LIBRARY
 	make_fractions_and_group(en);
 	if (factor_int_flag) {
-		factor_int_sub(en);
+		factor_int_equation(en);
 	}
 	free_result_str();
+#if	1	/* Set this to 1 to allow display2d to decide library output mode. */
+	if (display2d) {
+		result_str = flist_equation_string(en);
+		if (result_str == NULL)
+			result_str = list_equation(en, false);
+	} else {
+		result_str = list_equation(en, false);
+	}
+#else	/* For feeding command output to the next command's input only. */
 	result_str = list_equation(en, false);
+#endif
 	result_en = en;
 	if (gfp == stdout) {
 		return(result_str != NULL);
@@ -792,7 +1039,7 @@ int	en;	/* equation space number the result is in */
  * in the symbolic math library.
  */
 void
-free_result_str()
+free_result_str(void)
 {
 	if (result_str) {
 		free(result_str);
@@ -814,7 +1061,8 @@ char	*cp;
 /*
  * Process an equation number range given in text string "*cpp".
  * Skip past all spaces and update "*cpp" to point to the next argument if successful.
- * If no equation number or range is given, assume the current equation is wanted.
+ * If no equation number or range is given, or it is invalid, assume the current equation is wanted and
+ * don't skip anything.
  *
  * Return true if successful,
  * with the starting equation number in "*ip"
@@ -827,8 +1075,9 @@ int	*ip, *jp;
 {
 	int	i;
 	char	*cp;
+	int	rv;
 
-	cp = skip_space(*cpp);
+	cp = skip_comma_space(*cpp);
 	if (is_all(cp)) {
 		cp = skip_param(cp);
 		*ip = 0;
@@ -836,31 +1085,55 @@ int	*ip, *jp;
 		while (*jp > 0 && n_lhs[*jp] == 0)
 			(*jp)--;
 	} else {
+		if (*cp == '0') {
+			goto use_current;
+		}
 		if (isdigit(*cp)) {
 			*ip = strtol(cp, &cp, 10) - 1;
 		} else {
 			*ip = cur_equation;
 		}
-		if (*ip < 0 || *ip >= n_equations) {
-			error(_("Invalid equation number."));
-			return false;
-		}
 		if (*cp != '-') {
-			if (not_defined(*ip)) {
-				return false;
+			if (*cp == '\0' || *cp == ',' || isspace(*cp)) {
+				if (not_defined(*ip)) {
+					return false;
+				}
+				*jp = *ip;
+				*cpp = skip_comma_space(cp);
+				return true;
+			} else {
+use_current:
+				*jp = *ip = cur_equation;
+#if	1
+				rv = !empty_equation_space(cur_equation);	/* don't display error message */
+				if (rv) {
+					debug_string(1, _("Defaulting to the current equation space."));
+				} else {
+					debug_string(1, _("Defaulting to current empty equation space."));
+				}
+#else
+				rv = !current_not_defined();		/* display an error message if error */
+				if (rv) {
+					debug_string(1, _("Defaulting to the current equation space."));
+				}
+#endif
+				return rv;
 			}
-			*jp = *ip;
-			*cpp = skip_space(cp);
-			return true;
 		}
 		(cp)++;
+		if (*cp == '0') {
+			goto use_current;
+		}
 		if (isdigit(*cp)) {
 			*jp = strtol(cp, &cp, 10) - 1;
 		} else {
 			*jp = cur_equation;
 		}
-		if (*jp < 0 || *jp >= n_equations) {
-			error(_("Invalid equation number."));
+		if (*cp && !isspace(*cp)) {
+			goto use_current;
+		}
+		if (*ip < 0 || *ip >= N_EQUATIONS || *jp < 0 || *jp >= N_EQUATIONS) {
+			error(_("Invalid equation number (out of range)."));
 			return false;
 		}
 		if (*jp < *ip) {
@@ -869,14 +1142,14 @@ int	*ip, *jp;
 			*jp = i;
 		}
 	}
-	cp = skip_space(cp);
+	cp = skip_comma_space(cp);
 	for (i = *ip; i <= *jp; i++) {
 		if (n_lhs[i] > 0) {
 			*cpp = cp;
 			return true;
 		}
 	}
-	error(_("No equations defined in specified range."));
+	error(_("No expressions defined in specified range."));
 	return false;
 }
 
@@ -885,15 +1158,16 @@ int	*ip, *jp;
  *
  * Returns true if any non-space characters are encountered before the end of the string
  * and an error message is printed.
- * Otherwise just returns false.
+ * Otherwise just returns false indicating everything is OK.
  */
 int
 extra_characters(cp)
 char	*cp;	/* command line string */
 {
 	if (cp) {
-		cp = skip_space(cp);
+		cp = skip_comma_space(cp);
 		if (*cp) {
+			printf(_("\nError: \"%s\" not required on input line.\n"), cp);
 			error(_("Extra characters or unrecognized argument."));
 			return true;
 		}
@@ -933,9 +1207,24 @@ char	*cp;	/* character pointer */
 	return cp;
 }
 
+/*
+ * Skip over a possible comma and space characters.
+ */
+char *
+skip_comma_space(cp)
+char	*cp;	/* character pointer */
+{
+	if (cp) {
+		cp = skip_space(cp);
+		if (*cp == ',')
+			cp = skip_space(cp + 1);
+	}
+	return cp;
+}
+
 /*
  * Enhanced decimal strtol().
- * Skips trailing spaces.
+ * Skips trailing spaces or commas.
  */
 long
 decstrtol(cp, cpp)
@@ -944,8 +1233,8 @@ char	*cp, **cpp;
 	long	l;
 
 	l = strtol(cp, cpp, 10);
-	if (cpp && cp != *cpp) {
-		*cpp = skip_space(*cpp);
+	if (cpp && *cpp && cp != *cpp) {
+		*cpp = skip_comma_space(*cpp);
 	}
 	return l;
 }
@@ -962,7 +1251,7 @@ int	ch;
 
 /*
  * Skip over the current parameter in a Mathomatic command line string.
- * Parameters are usually separated with spaces.
+ * Parameters are usually separated with spaces or a comma or equals sign.
  *
  * Returns a string (character pointer) to the next parameter.
  */
@@ -975,7 +1264,7 @@ char	*cp;
 			cp++;
 		}
 		cp = skip_space(cp);
-		if (isdelimiter(*cp)) {
+		if (*cp && isdelimiter(*cp)) {
 			cp = skip_space(cp + 1);
 		}
 	}
@@ -983,9 +1272,8 @@ char	*cp;
 }
 
 /*
- * Compare strings up to the end or the first space or parameter delimiter.
- * Must be an exact match.
- * Compare is alphabetic case insensitive.
+ * Compare strings up to the end or the first space or parameter delimiter,
+ * ignoring alphabetic case.
  *
  * Returns zero on exact match, otherwise non-zero if strings are different.
  */
@@ -995,6 +1283,10 @@ char	*cp1, *cp2;
 {
 	char	*cp1a, *cp2a;
 
+#if	DEBUG
+	if (cp1 == NULL || cp2 == NULL)
+		error_bug("NULL pointer passed to strcmp_tospace().");
+#endif
 	for (cp1a = cp1; *cp1a && !isdelimiter(*cp1a); cp1a++)
 		;
 	for (cp2a = cp2; *cp2a && !isdelimiter(*cp2a); cp2a++)
@@ -1146,6 +1438,32 @@ int		n1;	/* expression length */
 	return true;
 }
 
+/*
+ * Test if expression contains an absolute value.
+ * Return true if it does.
+ */
+int
+exp_is_absolute(p1, n1)
+token_type	*p1;	/* expression pointer */
+int		n1;	/* expression length */
+{
+	int	i;
+	int	level;
+
+	for (i = n1 - 2; i > 2; i -= 2) {
+		if (p1[i].token.operatr != POWER)
+			continue;
+		level = p1[i].level;
+		if (p1[i+1].level == level && p1[i+1].kind == CONSTANT && fmod(p1[i+1].token.constant, 1.0) != 0.0) {
+			level++;
+			if (p1[i-2].token.operatr == POWER && p1[i-2].level == level && p1[i-1].level == level && p1[i-1].kind == CONSTANT) {
+				return true;
+			}
+		}
+	}
+	return false;
+}
+
 /*
  * Check for division by zero.
  *
@@ -1161,3 +1479,75 @@ double	denominator;
 	}
 	return false;
 }
+
+#if	CYGWIN || MINGW
+/*
+ * dirname(3) function for Microsoft Windows.
+ * dirname(3) strips the non-directory suffix from a filename.
+ */
+char *
+dirname_win(cp)
+char	*cp;	/* string containing filename to modify */
+{
+	int	i;
+
+	if (cp == NULL)
+		return(".");
+	i = strlen(cp);
+	while (i >= 0 && cp[i] != '\\' && cp[i] != '/')
+		i--;
+	if (i < 0)
+		return(".");
+	cp[i] = '\0';
+	return(cp);
+}
+#endif
+
+#if	!SECURE
+/*
+ * Load set options from startup file "~/.mathomaticrc".
+ *
+ * Return false if there was an error reading the startup file,
+ * otherwise return true.
+ */
+int
+load_rc(void)
+{
+	FILE	*fp = NULL;
+	char	buf[MAX_CMD_LEN];
+	char	*cp;
+	int	rv = true;
+
+	cp = getenv("HOME");
+	if (cp) {
+		snprintf(rc_file, sizeof(rc_file), "%s/%s", cp, ".mathomaticrc");
+		fp = fopen(rc_file, "r");
+	}
+#if	CYGWIN || MINGW
+	if (fp == NULL && cp) {
+		snprintf(rc_file, sizeof(rc_file), "%s/%s", cp, "mathomatic.rc");
+		fp = fopen(rc_file, "r");
+	}
+	if (fp == NULL && dir_path) {
+		snprintf(rc_file, sizeof(rc_file), "%s/%s", dir_path, "mathomatic.rc");
+		fp = fopen(rc_file, "r");
+	}
+#endif
+	if (fp == NULL) {
+		return true;
+	}
+	if (!quiet_mode && !eoption) {
+		printf(_("Loading startup set options from \"%s\".\n"), rc_file);
+	}
+	while ((cp = fgets(buf, sizeof(buf), fp)) != NULL) {
+		set_error_level(cp);
+		if (!set_options(cp, true))
+			rv = false;
+	}
+	if (fclose(fp)) {
+		rv = false;
+		perror(rc_file);
+	}
+	return rv;
+}
+#endif
