@@ -1,27 +1,30 @@
 /*
  * Mathomatic symbolic differentiation routines and related commands.
  *
- * Copyright (C) 1987-2010 George Gesslein II.
+ * Copyright (C) 1987-2012 George Gesslein II.
  
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public 
-    License as published by the Free Software Foundation; either 
-    version 2.1 of the License, or (at your option) any later version.
+This library is free software; you can redistribute it and/or
+modify it under the terms of the GNU Lesser General Public
+License as published by the Free Software Foundation; either
+version 2.1 of the License, or (at your option) any later version.
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of 
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
- 
-    You should have received a copy of the GNU Lesser General Public 
-    License along with this library; if not, write to the Free Software 
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+This library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+Lesser General Public License for more details.
+
+You should have received a copy of the GNU Lesser General Public
+License along with this library; if not, write to the Free Software
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+
+The chief copyright holder can be contacted at gesslein@mathomatic.org, or
+George Gesslein II, P.O. Box 224, Lansing, NY  14882-0224  USA.
  
  */
 
 #include "includes.h"
 
-static int	d_recurse();
+static int d_recurse(token_type *equation, int *np, int loc, int level, long v);
 
 /*
  * Compute the derivative of an equation side, with respect to variable "v",
@@ -350,53 +353,63 @@ char	*cp;
 		if (*cp) {
 			order = decstrtol(cp, &cp);
 		}
-		if (*cp || order <= 0) {
+		if (order <= 0) {
 			error(_("The order must be a positive integer."));
 			return false;
 		}
+		if (extra_characters(cp))
+			return false;
 	}
 	if (no_vars(source, *nps, &v)) {
-#if	!SILENT
-		error(_("Current expression contains no variables; the derivative would be zero."));
-		return false;
-#endif
+		warning(_("Current expression contains no variables; the derivative will be zero."));
+	} else {
+		if (v == 0) {
+			if (!prompt_var(&v)) {
+				return false;
+			}
+		}
+		if (v && v != MATCH_ANY && !found_var(source, *nps, v)) {
+			warning(_("Specified variable not found; the derivative will be zero."));
+		}
 	}
 	if (v == 0) {
-		if (!prompt_var(&v)) {
-			return false;
-		}
-	}
-#if	!SILENT
-	if (v != MATCH_ANY && !found_var(source, *nps, v)) {
-		error(_("Variable not found; the derivative would be zero."));
+		error(_("No differentiation variable specified."));
 		return false;
 	}
-	if (debug_level >= 0) {
-		list_var(v, 0);
-		if (n_rhs[cur_equation]) {
-			printf(_("Differentiating the RHS with respect to (%s)"), var_str);
-		} else {
-			printf(_("Differentiating with respect to (%s)"), var_str);
-		}
-		if (order != 1) {
-			printf(_(" %ld times"), order);
-		}
-		if (simplify_flag) {
-			printf(_(" and simplifying"));
-		}
-		printf("...\n");
+#if	!SILENT
+	list_var(v, 0);
+	if (n_rhs[cur_equation]) {
+		fprintf(gfp, _("Differentiating the RHS with respect to %s"), var_str);
+	} else {
+		fprintf(gfp, _("Differentiating with respect to %s"), var_str);
 	}
+	if (order != 1) {
+		fprintf(gfp, _(" %ld times"), order);
+	}
+	if (simplify_flag) {
+		fprintf(gfp, _(" and simplifying"));
+	} else {
+		fprintf(gfp, _(" and not simplifying"));
+	}
+	fprintf(gfp, "...\n");
 #endif
 	blt(dest, source, *nps * sizeof(token_type));
 	n1 = *nps;
 /* do the actual differentiating and simplifying: */
 	for (l1 = 0; l1 < order; l1++) {
+		if (order != 1) {
+			if (n1 == 1 && dest[0].kind == CONSTANT && dest[0].token.constant == 0.0) {
+				fprintf(gfp, _("0 reached after %ld derivatives taken.\n"), l1);
+				order = l1;
+				break;
+			}
+		}
 		if (!differentiate(dest, &n1, v)) {
 			error(_("Differentiation failed."));
 			return false;
 		}
 		if (simplify_flag) {
-			simpa_side(dest, &n1, true, false);
+			simpa_repeat_side(dest, &n1, true, false);
 		} else {
 			elim_loop(dest, &n1);
 		}
@@ -411,7 +424,9 @@ char	*cp;
 				var_str[len++] = '\'';
 			}
 			var_str[len] = '\0';
-			parse_var(&lhs[i][0].token.variable, var_str);
+			if (l1 == order) {
+				parse_var(&lhs[i][0].token.variable, var_str);
+			}
 		}
 	}
 	cur_equation = i;
@@ -457,11 +472,14 @@ char	*cp;
 		if (*cp) {
 			order = decstrtol(cp, &cp);
 		}
-		if (*cp || order <= 0) {
+		if (order <= 0) {
 			error(_("The order must be a positive integer."));
 			return false;
 		}
+		if (extra_characters(cp))
+			return false;
 	}
+	show_usage = false;
 	if (no_vars(source, n, &v)) {
 		error(_("Current expression contains no variables."));
 		return false;
@@ -472,7 +490,7 @@ char	*cp;
 		}
 	}
 	if (!found_var(source, n, v)) {
-		error(_("Variable not found; the derivative would be zero."));
+		error(_("Specified variable not found; the derivative would be zero."));
 		return false;
 	}
 	blt(rhs[i], source, n * sizeof(token_type));
@@ -482,7 +500,7 @@ char	*cp;
 			error(_("Differentiation failed."));
 			return false;
 		}
-		simpa_side(rhs[i], &n, true, false);
+		simpa_repeat_side(rhs[i], &n, true, false);
 	}
 	if (!found_var(rhs[i], n, v)) {
 		error(_("There are no solutions."));
@@ -501,7 +519,7 @@ char	*cp;
 		error(_("Solve failed."));
 		return false;
 	}
-	simpa_side(rhs[i], &n_rhs[i], false, false);
+	simpa_repeat_side(rhs[i], &n_rhs[i], false, false);
 	return return_result(cur_equation);
 }
 
@@ -517,7 +535,7 @@ char	*cp;
 	int		level;
 	long		l1, n, order = -1L;
 	double		d;
-	char		*cp_start, *cp1, buf[MAX_CMD_LEN];
+	char		*cp_start, *cp1 = NULL, buf[MAX_CMD_LEN];
 	int		our;
 	int		our_nlhs, our_nrhs;
 	token_type	*ep, *source, *dest;
@@ -540,6 +558,7 @@ char	*cp;
 	n_lhs[i] = 0;
         if (our < 0) {
                 error(_("Out of free equation spaces."));
+		show_usage = false;
 		return false;
 	}
 	if (n_rhs[cur_equation]) {
@@ -567,6 +586,7 @@ char	*cp;
 		}
 		cp = cp1;
 	}
+	show_usage = false;
 	no_vars(source, *nps, &v);
 	if (v == 0) {
 		if (!prompt_var(&v)) {
@@ -574,7 +594,7 @@ char	*cp;
 		}
 	}
 	if (!found_var(source, *nps, v)) {
-		warning(_("Variable not found."));
+		warning(_("Specified differentiation variable not found; the derivative will be 0."));
 	}
 	blt(rhs[our], source, *nps * sizeof(token_type));
 	our_nrhs = *nps;
@@ -586,7 +606,9 @@ char	*cp;
 	}
 	if (*cp) {
 		input_column += (cp - cp_start);
-		if ((cp = parse_expr(lhs[our], &our_nlhs, cp)) == NULL || our_nlhs <= 0) {
+		cp = parse_expr(lhs[our], &our_nlhs, cp, true);
+		if (cp == NULL || extra_characters(cp) || our_nlhs <= 0) {
+			show_usage = true;
 			return false;
 		}
 	} else {
@@ -594,7 +616,7 @@ char	*cp;
 		list_var(v, 0);
 		printf(_("Taylor series expansion around %s = point.\n"), var_str);
 #endif
-		my_strlcpy(prompt_str, _("Enter point: "), sizeof(prompt_str));
+		my_strlcpy(prompt_str, _("Enter point (expression; usually 0): "), sizeof(prompt_str));
 		if (!get_expr(lhs[our], &our_nlhs)) {
 			return false;
 		}
@@ -604,27 +626,32 @@ char	*cp;
 		if ((cp1 = get_string(buf, sizeof(buf))) == NULL)
 			return false;
 		if (*cp1) {
+			cp = NULL;
 			order = decstrtol(cp1, &cp);
-			if (*cp || order < 0) {
+			if (cp == NULL || *cp || order < 0) {
 				error(_("Positive integer required for order."));
 				return false;
 			}
 		} else {
-			order = LONG_MAX - 1L;
+			order = LONG_MAX - 1;
 #if	!SILENT
 			printf(_("Derivatives will be taken until they reach zero...\n"));
 #endif
 		}
 	}
 #if	!SILENT
-	printf(_("Computing the Taylor series"));
+	fprintf(gfp, _("Taylor series"));
 	if (n_rhs[cur_equation]) {
-		printf(_(" of the RHS"));
+		fprintf(gfp, _(" of the RHS"));
 	}
+	list_var(v, 0);
+	fprintf(gfp, _(" with respect to %s"), var_str);
 	if (simplify_flag) {
-		printf(_(" and simplifying"));
+		fprintf(gfp, _(", simplified"));
+	} else {
+		fprintf(gfp, _(", not simplified"));
 	}
-	printf("...\n");
+	fprintf(gfp, "...\n");
 #endif
 	n = 0;
 	i1 = 0;
@@ -704,9 +731,9 @@ loop_again:
 				return false;
 			}
 		}
-		symb_flag = symblify;
-		simpa_side(rhs[our], &our_nrhs, true, true);
-		symb_flag = false;
+/*		symb_flag = symblify; */
+		simpa_repeat_side(rhs[our], &our_nrhs, true, false /* was true */);
+/*		symb_flag = false; */
 		if (our_nrhs != 1 || rhs[our][0].kind != CONSTANT || rhs[our][0].token.constant != 0.0) {
 			i1 = n1;
 			if ((i1 + 1 + our_nrhs) > n_tokens)
@@ -724,7 +751,7 @@ loop_again:
 		}
 	}
 #if	!SILENT
-	printf(_("%ld derivative%s applied.\n"), n, (n == 1) ? "" : "s");
+	fprintf(gfp, _("%ld non-zero derivative%s applied.\n"), n, (n == 1) ? "" : "s");
 #endif
 	if (n_rhs[cur_equation]) {
 		n_lhs[i] = n_lhs[cur_equation];
@@ -758,7 +785,7 @@ char	*cp;
 		n_lhs[cur_equation] = 1;
 		lhs[cur_equation][0].level = 1;
 		lhs[cur_equation][0].kind = VARIABLE;
-		parse_var(&lhs[cur_equation][0].token.variable, "answer");
+		parse_var(&lhs[cur_equation][0].token.variable, "limit");
 	}
 	if (!solved_equation(cur_equation)) {
 		error(_("The current equation is not solved for a variable."));
@@ -772,9 +799,10 @@ char	*cp;
 			return false;
 		}
 	}
+	show_usage = false;
 	if (no_vars(rhs[cur_equation], n_rhs[cur_equation], &v)) {
-		error(_("Current expression contains no variables."));
-		return false;
+		warning(_("Current expression contains no variables; that is the answer."));
+		return return_result(cur_equation);
 	}
 	if (v == 0) {
 		if (!prompt_var(&v)) {
@@ -782,30 +810,40 @@ char	*cp;
 		}
 	}
 	if (!found_var(rhs[cur_equation], n_rhs[cur_equation], v)) {
-		error(_("Variable not found."));
-		return false;
+		warning(_("Limit variable not found; answer is original expression."));
+		return return_result(cur_equation);
 	}
 	if (*cp == '=') {
 		cp = skip_space(cp + 1);
 	}
 	if (*cp) {
 		input_column += (cp - cp_start);
-		if ((cp = parse_expr(tes, &n_tes, cp)) == NULL || n_tes <= 0) {
+		cp = parse_expr(tes, &n_tes, cp, true);
+		if (cp == NULL || extra_characters(cp) || n_tes <= 0) {
+			show_usage = true;
 			return false;
 		}
 	} else {
 		list_var(v, 0);
-		snprintf(prompt_str, sizeof(prompt_str), _("as (%s) goes to: "), var_str);
+		snprintf(prompt_str, sizeof(prompt_str), _("as %s goes to: "), var_str);
 		if (!get_expr(tes, &n_tes)) {
 			return false;
 		}
 	}
+	simp_loop(tes, &n_tes);
+#if	!SILENT
+	if (debug_level >= 0) {
+		list_var(v, 0);
+		fprintf(gfp, _("Taking the limit as %s goes to "), var_str);
+		list_proc(tes, n_tes, false);
+		fprintf(gfp, "\n");
+	}
+#endif
 /* copy the current equation to a new equation space, then simplify and work on the copy: */
 	copy_espace(cur_equation, i);
 	simpa_side(rhs[i], &n_rhs[i], false, false);
 
 /* see if the limit expression is positive infinity: */
-	simp_loop(tes, &n_tes);
 	if (n_tes == 1 && tes[0].kind == CONSTANT && tes[0].token.constant == INFINITY) {
 /* To take the limit to positive infinity, */
 /* replace infinity with zero and replace the limit variable with its reciprocal: */
@@ -823,7 +861,7 @@ char	*cp;
 	}
 
 /* General limit taking, solve for the limit variable: */
-	debug_string(0, "Solving...");
+	debug_string(0, _("Solving..."));
 	want.level = 1;
 	want.kind = VARIABLE;
 	want.token.variable = v;
