@@ -1,29 +1,32 @@
 /*
  * Mathomatic integration routines and commands.
  *
- * Copyright (C) 1987-2010 George Gesslein II.
+ * Copyright (C) 1987-2012 George Gesslein II.
  
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public 
-    License as published by the Free Software Foundation; either 
-    version 2.1 of the License, or (at your option) any later version.
+This library is free software; you can redistribute it and/or
+modify it under the terms of the GNU Lesser General Public
+License as published by the Free Software Foundation; either
+version 2.1 of the License, or (at your option) any later version.
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of 
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
- 
-    You should have received a copy of the GNU Lesser General Public 
-    License along with this library; if not, write to the Free Software 
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+This library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+Lesser General Public License for more details.
+
+You should have received a copy of the GNU Lesser General Public
+License along with this library; if not, write to the Free Software
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+
+The chief copyright holder can be contacted at gesslein@mathomatic.org, or
+George Gesslein II, P.O. Box 224, Lansing, NY  14882-0224  USA.
  
  */
 
 #include "includes.h"
 
-static int	integrate_sub();
-static int	laplace_sub();
-static int	inv_laplace_sub();
+static int integrate_sub(token_type *equation, int *np, int loc, int eloc, long v);
+static int laplace_sub(token_type *equation, int *np, int loc, int eloc, long v);
+static int inv_laplace_sub(token_type *equation, int *np, int loc, int eloc, long v);
 
 static int	constant_var_number = 1;	/* makes unique numbers for the constant of integration */
 
@@ -82,7 +85,7 @@ int_dispatch(equation, np, v, func)
 token_type	*equation;	/* pointer to beginning of equation side to integrate */
 int		*np;		/* pointer to length of equation side */
 long		v;		/* integration variable */
-int		(*func)();	/* integration function to call for each term */
+int		(*func)(token_type *equation, int *np, int loc, int eloc, long v);	/* integration function to call for each term */
 {
 	int	i, j;
 
@@ -250,12 +253,15 @@ char	*cp;
 	int		n1, n2, *nps, *np;
 	int		definite_flag = false, constant_flag = false, solved;
 	double		integrate_order = 1.0;
-	char		var_name_buf[MAX_VAR_LEN];
+	char		var_name_buf[MAX_VAR_LEN], *cp_start;
 	long		l1;
 
+	cp_start = cp;
 	if (current_not_defined()) {
 		return false;
 	}
+	n_tlhs = 0;
+	n_trhs = 0;
 	solved = solved_equation(cur_equation);
 	i = next_espace();
 	for (;; cp = skip_param(cp)) {
@@ -294,11 +300,30 @@ char	*cp;
 		if (*cp) {
 			integrate_order = strtod(cp, &cp);
 		}
-		if (*cp || integrate_order <= 0 || fmod(integrate_order, 1.0) != 0.0) {
+		if (!isfinite(integrate_order) || integrate_order <= 0 || fmod(integrate_order, 1.0) != 0.0) {
 			error(_("The order must be a positive integer."));
 			return false;
 		}
 	}
+
+	if (*cp) {
+		cp = skip_comma_space(cp);
+		input_column += (cp - cp_start);
+		cp = parse_expr(tlhs, &n_tlhs, cp, false);
+		if (cp == NULL || n_tlhs <= 0) {
+			return false;
+		}
+	}
+	if (*cp) {
+		cp_start = cp;
+		cp = skip_comma_space(cp);
+		input_column += (cp - cp_start);
+		cp = parse_expr(trhs, &n_trhs, cp, false);
+		if (cp == NULL || extra_characters(cp) || n_trhs <= 0) {
+			return false;
+		}
+	}
+	show_usage = false;
 	if (v == 0) {
 		if (!prompt_var(&v)) {
 			return false;
@@ -308,14 +333,14 @@ char	*cp;
 	if (debug_level >= 0) {
 		list_var(v, 0);
 		if (n_rhs[cur_equation]) {
-			printf(_("Integrating the RHS with respect to (%s)"), var_str);
+			fprintf(gfp, _("Integrating the RHS with respect to %s"), var_str);
 		} else {
-			printf(_("Integrating with respect to (%s)"), var_str);
+			fprintf(gfp, _("Integrating with respect to %s"), var_str);
 		}
 		if (integrate_order != 1.0) {
-			printf(_(" %.*g times"), precision, integrate_order);
+			fprintf(gfp, _(" %.*g times"), precision, integrate_order);
 		}
-		printf(_(" and simplifying...\n"));
+		fprintf(gfp, _(" and simplifying...\n"));
 	}
 #endif
 	partial_flag = false;
@@ -355,13 +380,17 @@ char	*cp;
 		simp_loop(dest, &n1);
 	}
 	if (definite_flag) {
-		my_strlcpy(prompt_str, _("Enter lower bound: "), sizeof(prompt_str));
-		if (!get_expr(tlhs, &n_tlhs)) {
-			return false;
+		if (n_tlhs == 0) {
+			my_strlcpy(prompt_str, _("Enter lower bound: "), sizeof(prompt_str));
+			if (!get_expr(tlhs, &n_tlhs)) {
+				return false;
+			}
 		}
-		my_strlcpy(prompt_str, _("Enter upper bound: "), sizeof(prompt_str));
-		if (!get_expr(trhs, &n_trhs)) {
-			return false;
+		if (n_trhs == 0) {
+			my_strlcpy(prompt_str, _("Enter upper bound: "), sizeof(prompt_str));
+			if (!get_expr(trhs, &n_trhs)) {
+				return false;
+			}
 		}
 		blt(scratch, dest, n1 * sizeof(token_type));
 		n2 = n1;
@@ -591,13 +620,15 @@ char	*cp;
 		if (cp == NULL) {
 			return false;
 		}
-	} else {
-		if (!prompt_var(&v)) {
+		if (extra_characters(cp)) {
 			return false;
 		}
 	}
-	if (extra_characters(cp)) {
-		return false;
+	show_usage = false;
+	if (v == 0) {
+		if (!prompt_var(&v)) {
+			return false;
+		}
 	}
 	partial_flag = false;
 	uf_simp(source, nps);
@@ -607,12 +638,12 @@ char	*cp;
 	n1 = *nps;
 	if (inverse_flag) {
 		if (!poly_in_v(dest, n1, v, true) || !int_dispatch(dest, &n1, v, inv_laplace_sub)) {
-			error(_("Inverse Laplace failed."));
+			error(_("Inverse Laplace transformation failed."));
 			return false;
 		}
 	} else {
 		if (!poly_in_v(dest, n1, v, false) || !int_dispatch(dest, &n1, v, laplace_sub)) {
-			error(_("Laplace failed, not a polynomial."));
+			error(_("Laplace transformation failed, not a polynomial."));
 			return false;
 		}
 	}
@@ -637,7 +668,7 @@ int
 nintegrate_cmd(cp)
 char	*cp;
 {
-	long		v = 0;
+	long		v = 0;			/* Mathomatic variable */
 	int		i, j, k, i1, i2;
 	int		len;
 	int		level;
@@ -646,10 +677,14 @@ char	*cp;
 	int		trap_flag, singularity, solved;
 	token_type	*ep, *source, *dest;
 	int		n1, *nps, *np;
+	char		*cp_start;
 
+	cp_start = cp;
 	if (current_not_defined()) {
 		return false;
 	}
+	n_tlhs = 0;
+	n_trhs = 0;
 	solved = solved_equation(cur_equation);
 	i = next_espace();
 	if (n_rhs[cur_equation]) {
@@ -675,14 +710,43 @@ char	*cp;
 		if (*cp) {
 			iterations = decstrtol(cp, &cp);
 		}
-		if (*cp || iterations <= 0 || (iterations % 2) != 0) {
+		if (iterations <= 0 || (iterations % 2) != 0) {
 			error(_("Number of partitions must be a positive, even integer."));
 			return false;
 		}
-	} else {
-		if (!prompt_var(&v))
-			return false;
 	}
+	if (*cp) {
+		input_column += (cp - cp_start);
+		cp = parse_expr(tlhs, &n_tlhs, cp, false);
+		if (cp == NULL || n_tlhs <= 0) {
+			return false;
+		}
+	}
+	if (*cp) {
+		cp_start = cp;
+		cp = skip_comma_space(cp);
+		input_column += (cp - cp_start);
+		cp = parse_expr(trhs, &n_trhs, cp, false);
+		if (cp == NULL || extra_characters(cp) || n_trhs <= 0) {
+			return false;
+		}
+	}
+	show_usage = false;
+	if (v == 0) {
+		if (!prompt_var(&v)) {
+			return false;
+		}
+	}
+#if	!SILENT
+	if (debug_level >= 0) {
+		list_var(v, 0);
+		if (n_rhs[cur_equation]) {
+			fprintf(gfp, _("Numerically integrating the RHS with respect to %s...\n"), var_str);
+		} else {
+			fprintf(gfp, _("Numerically integrating with respect to %s...\n"), var_str);
+		}
+	}
+#endif
 	singularity = false;
 	for (j = 1; j < *nps; j += 2) {
 		if (source[j].token.operatr == DIVIDE) {
@@ -690,43 +754,47 @@ char	*cp;
 				if (source[k].kind == VARIABLE && source[k].token.variable == v) {
 					singularity = true;
 				}
+				if (source[k].kind == OPERATOR && source[k].level == source[j].level)
+					break;
 			}
 		}
 	}
 	if (singularity) {
 		warning(_("Singularity detected, result of numerical integration might be wrong."));
 	}
-	my_strlcpy(prompt_str, _("Enter lower bound: "), sizeof(prompt_str));
-	if (!get_expr(tlhs, &n_tlhs)) {
-		return false;
+	if (n_tlhs == 0) {
+		my_strlcpy(prompt_str, _("Enter lower bound: "), sizeof(prompt_str));
+		if (!get_expr(tlhs, &n_tlhs)) {
+			return false;
+		}
 	}
 	subst_constants(tlhs, &n_tlhs);
+	simp_loop(tlhs, &n_tlhs);
 	if (exp_contains_infinity(tlhs, n_tlhs)) {
-		error(_("Error: Bound contains infinity."));
+		error(_("Not computable because: Lower bound contains infinity or NaN."));
 		return false;
 	}
-	my_strlcpy(prompt_str, _("Enter upper bound: "), sizeof(prompt_str));
-	if (!get_expr(trhs, &n_trhs)) {
-		return false;
+	if (n_trhs == 0) {
+		my_strlcpy(prompt_str, _("Enter upper bound: "), sizeof(prompt_str));
+		if (!get_expr(trhs, &n_trhs)) {
+			return false;
+		}
 	}
 	subst_constants(trhs, &n_trhs);
+	simp_loop(trhs, &n_trhs);
 	if (exp_contains_infinity(trhs, n_trhs)) {
-		error(_("Error: Bound contains infinity."));
+		error(_("Not computable because: Upper bound contains infinity or NaN."));
 		return false;
 	}
 	if ((n_tlhs + n_trhs + 3) > n_tokens) {
 		error_huge();
 	}
 #if	!SILENT
-	if (n_rhs[cur_equation]) {
-		printf(_("Approximating the definite integral of the RHS\n"));
-	} else {
-		printf(_("Approximating the definite integral\n"));
-	}
+	fprintf(gfp, _("Approximating the definite integral\n"));
 	if (trap_flag) {
-		printf(_("using the trapezoid method (%d partitions)...\n"), iterations);
+		fprintf(gfp, _("using the trapezoid method (%d partitions)...\n"), iterations);
 	} else {
-		printf(_("using Simpson's rule (%d partitions)...\n"), iterations);
+		fprintf(gfp, _("using Simpson's rule (%d partitions)...\n"), iterations);
 	}
 #endif
 	subst_constants(source, nps);
@@ -837,7 +905,9 @@ char	*cp;
 		elim_loop(dest, &n1);
 		ufactor(dest, &n1);
 		simp_divide(dest, &n1);
+		factor_imaginary(dest, &n1);
 		approximate_roots = false;
+		side_debug(1, dest, n1);
 
 		if (exp_contains_infinity(dest, n1)) {
 			error(_("Integration failed because result contains infinity or NaN (a singularity)."));
@@ -854,7 +924,7 @@ char	*cp;
 			break;
 		default:
 			if ((n1 / 8) >= first_size) {
-				error(_("Integration failed."));
+				error(_("Result growing, integration failed."));
 				return false;
 			}
 			break;
@@ -887,15 +957,17 @@ char	*cp;
 	for (; k < n1; k++)
 		dest[k].level++;
 
-	/* simplify and approximate the result as before: */
+	/* simplify and approximate the result even more: */
 	approximate_roots = true;
-	elim_loop(dest, &n1);
-	ufactor(dest, &n1);
-	simp_divide(dest, &n1);
+	do {
+		elim_loop(dest, &n1);
+		ufactor(dest, &n1);
+		simp_divide(dest, &n1);
+	} while (factor_imaginary(dest, &n1));
 	approximate_roots = false;
 
 #if	!SILENT
-	printf(_("Numerical integration successful:\n"));
+	fprintf(gfp, _("Numerical integration successful:\n"));
 #endif
 	*np = n1;
 	if (n_rhs[cur_equation]) {
@@ -909,6 +981,5 @@ char	*cp;
 			parse_var(&lhs[i][0].token.variable, var_str);
 		}
 	}
-	cur_equation = i;
-	return return_result(cur_equation);
+	return return_result(i);
 }
